from .CoordinateSystemConverter import CoordinateSystemConverter
from .CommonCoordinateSystems import CartesianCoordinates3D, ZMatrixCoordinates
from ...Numputils import vec_norms, vec_angles, pts_dihedrals, dist_deriv, angle_deriv, dihed_deriv
import numpy as np
# this import gets bound at load time, so unfortunately PyCharm can't know just yet
# what properties its class will have and will try to claim that the files don't exist

class CartesianToZMatrixConverter(CoordinateSystemConverter):
    """
    A converter class for going from Cartesian coordinates to ZMatrix coordinates
    """

    @property
    def types(self):
        return (CartesianCoordinates3D, ZMatrixCoordinates)

    @staticmethod
    def get_dists(points, centers):
        return vec_norms(centers-points)
    @staticmethod
    def get_angles(lefts, centers, rights):
        # need to look up again what the convention is for which atom is the central one...
        v1s = centers-lefts
        v2s = centers-rights
        return vec_angles(v1s, v2s)[0]
    @staticmethod
    def get_diheds(points, centers, seconds, thirds):
        return pts_dihedrals(points, centers, seconds, thirds)

    def convert_many(self, coords, ordering=None, use_rad=True, return_derivs=False, **kw):
        """
        We'll implement this by having the ordering arg wrap around in coords?
        """
        if ordering is None:
            ordering = range(len(coords[0]))
        base_shape = coords.shape
        new_coords = np.reshape(coords, (np.prod(base_shape[:-1]),) + base_shape[-1:])
        new_coords, ops = self.convert(new_coords, ordering=ordering, use_rad=use_rad, return_derivs=return_derivs)
        single_coord_shape = (base_shape[-2]-1, new_coords.shape[-1])
        new_shape = base_shape[:-2] + single_coord_shape
        new_coords = np.reshape(new_coords, new_shape)
        if return_derivs:
            ders = ops['derivs']
            # we assume we get a list of derivatives?
            reshaped_ders = [None]*len(ders)
            for i, v in enumerate(ders):
                single_base_shape = (base_shape[-2], new_coords.shape[-1])
                ders_shape = coords.shape + single_base_shape*i + single_coord_shape
                v = v.reshape(ders_shape)
                reshaped_ders[i] = v
                ops['derivs'] = reshaped_ders
        return new_coords, ops

    def convert(self, coords, ordering=None, use_rad=True, return_derivs=False, **kw):
        """The ordering should be specified like:

        [
            [n1],
            [n2, n1]
            [n3, n1/n2, n1/n2]
            [n4, n1/n2/n3, n1/n2/n3, n1/n2/n3]
            [n5, ...]
            ...
        ]

        :param coords:    array of cartesian coordinates
        :type coords:     np.ndarray
        :param use_rad:   whether to user radians or not
        :type use_rad:    bool
        :param ordering:  optional ordering parameter for the z-matrix
        :type ordering:   None or tuple of ints or tuple of tuple of ints
        :param kw:        ignored key-word arguments
        :type kw:
        :return: z-matrix coords
        :rtype: np.ndarray
        """
        ncoords = len(coords)
        orig_ol = ZMatrixCoordinates.canonicalize_order_list(ncoords, ordering)
        ol = orig_ol
        nol = len(ol)
        ncol = len(ol[0])
        fsteps = ncoords / nol
        steps = int(fsteps)

        # print(">> c2z >> ordering:", ol)

        multiconfig = nol < ncoords
        if multiconfig:
            ol = ZMatrixCoordinates.tile_order_list(ol, ncoords)
            mc_ol = ol.copy()

        # we define an order map that we'll index into to get the new indices for a
        # given coordinate
        om = 1+np.argsort(ol[:, 0])

        # need to check against the cases of like 1, 2, 3 atom molecules
        # annoying but not hard
        if return_derivs:
            derivs = [
                np.zeros(coords.shape + (nol-1, 3)),
                np.zeros(coords.shape + (nol, 3) + (nol - 1, 3))
            ]
        if not multiconfig:
            ix = ol[1:, 0]
            jx = ol[1:, 1]
            dists = self.get_dists(coords[ix], coords[jx])
            if return_derivs:
                _dists, dist_derivs, dist_derivs_2 = dist_deriv(coords, ix, jx, order=2)
                drang = np.arange(len(ix))
                derivs[0][ix, :, drang, 0] = dist_derivs[0]
                derivs[0][jx, :, drang, 0] = dist_derivs[1]

                for i, x1 in enumerate([ix, jx]):
                    for j, x2 in enumerate([ix, jx]):
                        # print(i, j, x1, x2,
                        #       # dist_derivs_2[i, j][0, 0],
                        #       drang
                        #       )
                        derivs[1][x1, :, x2, :, drang, 0] = dist_derivs_2[i, j]

            if len(ol) > 2:
                ix = ol[2:, 0]
                jx = ol[2:, 1]
                kx = ol[2:, 2]
                angles = np.concatenate( (
                    [0], self.get_angles(coords[ix], coords[jx], coords[kx])
                ) )
                if not use_rad:
                    angles = np.rad2deg(angles)
                if return_derivs:
                    _angles, angle_derivs, angle_derivs_2 = angle_deriv(coords, jx, ix, kx, order=2)
                    drang = 1+np.arange(len(ix))
                    # print(">>>>", np.max(np.abs(angle_derivs)))
                    derivs[0][jx, :, drang, 1] = angle_derivs[0]
                    derivs[0][ix, :, drang, 1] = angle_derivs[1]
                    derivs[0][kx, :, drang, 1] = angle_derivs[2]

                    for i, x1 in enumerate([jx, ix, kx]):
                        for j, x2 in enumerate([jx, ix, kx]):
                            derivs[1][x1, :, x2, :, drang, 1] = angle_derivs_2[i, j]
            else:
                angles = np.array([0.])
            if len(ol) > 3:
                ix = ol[3:, 0]
                jx = ol[3:, 1]
                kx = ol[3:, 2]
                lx = ol[3:, 3]
                if ol.shape[1] == 5:
                    raise NotImplementedError("psi angles might be unnecessary")
                    ix = ix.copy()
                    jx = jx.copy()
                    kx = kx.copy()
                    lx = lx.copy()
                    fx = ol[3:, 4]
                    swap_pos = np.where(fx == 1)
                    swap_i = ix[swap_pos]
                    swap_j = jx[swap_pos]
                    swap_k = kx[swap_pos]
                    swap_l = lx[swap_pos]
                    ix[swap_pos] = swap_l
                    jx[swap_pos] = swap_i
                    kx[swap_pos] = swap_j
                    lx[swap_pos] = swap_k

                diheds = np.concatenate( (
                    [0, 0],
                    self.get_diheds(coords[ix], coords[jx], coords[kx], coords[lx])
                ) )
                if not use_rad:
                    diheds = np.rad2deg(diheds)
                if return_derivs:
                    _diheds, dihed_derivs, dihed_derivs_2 = dihed_deriv(coords, ix, jx, kx, lx, order=2)
                    drang = 2+np.arange(len(ix))
                    derivs[0][ix, :, drang, 2] = dihed_derivs[0]
                    derivs[0][jx, :, drang, 2] = dihed_derivs[1]
                    derivs[0][kx, :, drang, 2] = dihed_derivs[2]
                    derivs[0][lx, :, drang, 2] = dihed_derivs[3]

                    for i, x1 in enumerate([ix, jx, kx, lx]):
                        for j, x2 in enumerate([ix, jx, kx, lx]):
                            derivs[1][x1, :, x2, :, drang, 2] = dihed_derivs_2[i, j]
            else:
                diheds = np.array([0, 0])
            ol = ol[1:]

        else: # multiconfig

            # we do all of this stuff with masking operations in the multiconfiguration cases
            mask = np.repeat(True, ncoords)
            mask[np.arange(0, ncoords, nol)] = False
            ix = ol[mask, 0]
            jx = ol[mask, 1]
            dists = self.get_dists(coords[ix], coords[jx])
            if return_derivs:
                _, dist_derivs, dist_derivs_2 = dist_deriv(coords, ix, jx, order=2)
                drang = np.arange(nol-1)
                nreps = int(len(ix)/(nol-1))
                drang = np.broadcast_to(drang[np.newaxis], (nreps,) + drang.shape).flatten()
                derivs[0][ix, :, drang, 0] = dist_derivs[0]
                derivs[0][jx, :, drang, 0] = dist_derivs[1]

                for i, x1 in enumerate([ix, jx]):
                    for j, x2 in enumerate([ix, jx]):
                        derivs[1][x1, :, x2 % nol, :, drang, 0] = dist_derivs_2[i, j]

            if nol>2:
                # set up the mask to drop all of the first bits
                mask[np.arange(1, ncoords, nol)] = False
                ix = ol[mask, 0]
                jx = ol[mask, 1]
                kx = ol[mask, 2]
                angles = self.get_angles(coords[ix], coords[jx], coords[kx])
                angles = np.append(angles, np.zeros(steps))
                insert_pos = np.arange(0, ncoords-1*steps-1, nol-2)
                angles = np.insert(angles, insert_pos, 0)
                angles = angles[:ncoords-steps]
                if not use_rad:
                    angles = np.rad2deg(angles)
                if return_derivs:
                    # we might need to mess with the masks akin to the insert call...
                    _, angle_derivs, angle_derivs_2 = angle_deriv(coords, jx, ix, kx, order=2)
                    drang = 1+np.arange(nol-2)
                    nreps = int(len(ix)/(nol-2))
                    drang = np.broadcast_to(drang[np.newaxis], (nreps,) + drang.shape).flatten()
                    derivs[0][jx, :, drang, 1] = angle_derivs[0]
                    derivs[0][ix, :, drang, 1] = angle_derivs[1]
                    derivs[0][kx, :, drang, 1] = angle_derivs[2]

                    for i, x1 in enumerate([ix, jx, kx]):
                        for j, x2 in enumerate([ix, jx, kx]):
                            derivs[1][x1, :, x2 % nol, :, drang, 0] = angle_derivs_2[i, j]
            else:
                angles = np.zeros(ncoords-steps)

            if nol > 3:
                # set up mask to drop all of the second atom bits (wtf it means 'second')
                mask[np.arange(2, ncoords, nol)] = False
                ix = ol[mask, 0]
                jx = ol[mask, 1]
                kx = ol[mask, 2]
                lx = ol[mask, 3]
                if ol.shape[1] == 5:
                    raise ValueError("Unclear if there is a difference between tau and psi")
                    ix = ix.copy()
                    jx = jx.copy()
                    kx = kx.copy()
                    lx = lx.copy()
                    fx = ol[mask, 4]
                    swap_pos = np.where(fx == 1)
                    swap_i = ix[swap_pos]
                    swap_j = jx[swap_pos]
                    swap_k = kx[swap_pos]
                    swap_l = lx[swap_pos]
                    ix[swap_pos] = swap_l
                    jx[swap_pos] = swap_i
                    kx[swap_pos] = swap_j
                    lx[swap_pos] = swap_k
                # print(ol)

                diheds = self.get_diheds(coords[ix], coords[jx], coords[kx], coords[lx])
                # pad diheds to be the size of ncoords
                diheds = np.append(diheds, np.zeros(2*steps))

                # insert zeros where undefined
                diheds = np.insert(diheds, np.repeat(np.arange(0, ncoords-2*steps-1, nol-3), 2), 0)
                # take only as many as actually used
                diheds = diheds[:ncoords-steps]
                if not use_rad:
                    diheds = np.rad2deg(diheds)
                if return_derivs:
                    # Negative sign because my dihed_deriv code is for slightly different
                    # ordering than expected
                    _, dihed_derivs, dihed_derivs_2 = dihed_deriv(coords, ix, jx, kx, lx, order=2)
                    drang = 2+np.arange(nol-3)
                    nreps = int(len(ix)/(nol-3))
                    drang = np.broadcast_to(drang[np.newaxis], (nreps,) + drang.shape).flatten()
                    derivs[0][ix, :, drang, 2] = dihed_derivs[0]
                    derivs[0][jx, :, drang, 2] = dihed_derivs[1]
                    derivs[0][kx, :, drang, 2] = dihed_derivs[2]
                    derivs[0][lx, :, drang, 2] = dihed_derivs[3]

                    for i, x1 in enumerate([ix, jx, kx, lx]):
                        for j, x2 in enumerate([ix, jx, kx, lx]):
                            derivs[1][x1, :, x2 % nol, :, drang, 0] = dihed_derivs_2[i, j]

            else:
                diheds = np.zeros(ncoords-steps)

            # after the np.insert calls we have the right number of final elements, but too many
            # ol and om elements and they're generally too large
            # so we need to shift them down and mask out the elements we don't want
            mask = np.repeat(True, ncoords)
            mask[np.arange(0, ncoords, nol)] = False
            ol = np.reshape(ol[mask], (steps, nol-1, ncol))-np.reshape(np.arange(steps), (steps, 1, 1))
            ol = np.reshape(ol, (ncoords-steps, ncol))
            om = np.reshape(om[mask], (steps, nol-1))-nol*np.reshape(np.arange(steps), (steps, 1))-1
            om = np.reshape(om, (ncoords-steps,))

        final_coords = np.array(
            [
                dists, angles, diheds
            ]
        ).T

        if multiconfig:
            # figure out what to use for the axes
            origins = coords[mc_ol[1::nol,  1]]
            x_axes  = coords[mc_ol[1::nol,  0]] - origins # the first displacement vector
            y_axes  = coords[mc_ol[2::nol,  0]] - origins # the second displacement vector (just defines the x-y plane, not the real y-axis)
            axes = np.array([x_axes, y_axes]).transpose((1, 0, 2))
        else:
            origins = coords[ol[0, 1]]
            axes = np.array([coords[ol[0, 0]] - origins, coords[ol[1, 0]] - origins])

        ol = orig_ol
        om = om - 1
        if ncol == 5:
            ordering = np.array([
                        np.argsort(ol[:, 0]), om[ol[:, 1]], om[ol[:, 2]], om[ol[:, 3]], ol[:, 4]
                    ]).T
        else:
            ordering = np.array([
                    np.argsort(ol[:, 0]), om[ol[:, 1]], om[ol[:, 2]], om[ol[:, 3]]
                ]).T
        opts = dict(use_rad=use_rad, ordering=ordering, origins=origins, axes=axes)

        # if we're returning derivs, we also need to make sure that they're ordered the same way the other data is...
        if return_derivs:
            opts['derivs'] = derivs#[:1]

        return final_coords, opts

__converters__ = [ CartesianToZMatrixConverter() ]


"""
Provides access to the JHTML interface for building quality Jupyter interfaces
"""

__all__ = []
from .JHTML import *; from .JHTML import __all__ as exposed
__all__ += exposed
from .WidgetTools import *; from .WidgetTools import __all__ as exposed
__all__ += exposed
from .HTML import *; from .HTML import __all__ as exposed
__all__ += exposed
from .Bootstrap import *; from .Bootstrap import __all__ as exposed
__all__ += exposed
from .HTMLWidgets import *; from .HTMLWidgets import __all__ as exposed
__all__ += exposed
from .BootstrapWidgets import *; from .BootstrapWidgets import __all__ as exposed
__all__ += exposed


"""
A growing package of assorted functionality that finds use across many different packages, but doesn't attempt to
provide a single unified interface for doing certain types of projects.

All of the McUtils packages stand mostly on their own, but there will be little calls into one another here and there.

The more scientifically-focused `Psience` package makes significant use of `McUtils` as do various packages that have
been written over the years.
"""

import McUtils.Data
import McUtils.Parsers
import McUtils.GaussianInterface
import McUtils.Misc
import McUtils.Docs
import McUtils.Jupyter
import McUtils.Plots
import McUtils.Zachary
import McUtils.Coordinerds
import McUtils.ExternalPrograms
import McUtils.Extensions

__all__ = [
    "Parsers",
    "Extensions",
    "Plots",
    "ExternalPrograms",
    "Zachary",
    "Data",
    "Coordinerds",
    "GaussianInterface",
    "Numputils",
    "Scaffolding",
    "Parallelizers",
    "Jupyter",
    "Misc",
    "Formatters",
    "Docs"
]



"""
Provides a class that will walk through a set of objects & their children, as loaded into memory,
and call appropriate handlers for each
"""

import os, types, collections, abc
import sys, importlib
# from .Writers import *

__all__ = [
    "ObjectWalker",
    "ObjectHandler",
    "ObjectSpec"
]

class ObjectTree(dict):
    """
    Simple tree that stores the structure of the documentation
    """
class ObjectSpec(dict):
    required_keys = ['id']

class MethodDispatch(collections.OrderedDict):
    """
    Provides simple utility to dispatch methods based on types
    """

    def __init__(self, *args, default=None, **kwargs):
        self.default = default
        super().__init__(*args, **kwargs)
    class DispatchTests:
        def __init__(self, *tests):
            self.tests = tests
        def __hash__(self):
            return self.tests.__hash__()
        def __call__(self, obj):
            return all(self.test(t, obj) for t in self.tests)
        @classmethod
        def test(cls, k, obj):
            """
            Does the actual dispatch testing

            :param k:
            :type k:
            :param obj:
            :type obj:
            :return:
            :rtype:
            """
            if (
                    isinstance(k, type) or
                    isinstance(k, tuple) and all(isinstance(kk, type) for kk in k)
            ):
                return isinstance(obj, k)
            elif isinstance(k, str):
                return hasattr(obj, k)
            elif isinstance(k, tuple) and all(isinstance(kk, str) for kk in k):
                return any(hasattr(obj, kk) for kk in k)
            elif isinstance(k, tuple):
                return any(kk(obj) for kk in k)
            elif k is None:
                return True
            else:
                return k(obj)
    def method_dispatch(self, obj, *args, **kwargs):
        """
        A general-ish purpose type or duck-type method dispatcher.

        :param obj:
        :type obj:
        :param table:
        :type table:
        :return:
        :rtype:
        """

        for k, v in self.items():
            if isinstance(k, self.DispatchTests):
                matches = k(obj)
            else:
                matches = self.DispatchTests.test(k, obj)
            if matches:
                return v(obj, *args, **kwargs)

        if self.default is None:
            raise TypeError("object {} can't dispatch from table {}".format(
                obj, self
            ))
        else:
            return self.default(obj, *args, **kwargs)
    def __call__(self, obj, *args, **kwargs):
        return self.method_dispatch(obj, *args, **kwargs)
    def __setitem__(self, key, value):
        """
        :param key:
        :type key:
        :param value:
        :type value:
        :return:
        :rtype:
        """
        #TODO: make dispatch keys automatically feed into
        if isinstance(key, tuple):
            # make sure we're not just doing alternatives
            if not (
                    all(isinstance(k, str) for k in key) or
                    all(isinstance(k, type) for k in key) or
                    all(callable(k) for k in key)
            ):
                # then we do, basically, an 'and' operand
                key = self.DispatchTests(*key)
        super().__setitem__(key, value)

class ObjectHandler(metaclass=abc.ABCMeta):
    protected_fields = set()
    default_fields = {}
    def __init__(self,
                 obj,
                 *,
                 spec=None,
                 tree=None,
                 name=None,
                 parent=None,
                 walker:'ObjectWalker'=None,
                 extra_fields=None,
                 **kwargs
                 ):

        self.walker = walker
        self.obj = obj
        self._id = None
        self._name = name
        self._parent = parent
        self._pobj = None
        self._chobj = None
        self.spec = {} if spec is None else spec
        self.tree = tree

        if extra_fields is None:
            extra_fields = kwargs
        else:
            extra_fields = dict(kwargs, **extra_fields)

        self.extra_fields = extra_fields
        for k in self.default_fields.keys() - self.extra_fields.keys():
            self.extra_fields[k] = self.default_fields[k]
        for k in self.extra_fields.keys() & set(self.protected_fields):
            del self.extra_fields[k]


    def __getitem__(self, item):
        return self.resolve_key(item)
    def resolve_key(self, key, default=None):
        if key in self.spec:
            return self.spec.get(key, default)
        elif key in self.extra_fields:
            return self.extra_fields.get(key, default)
        else:
            return default

    @property
    def name(self):
        """
        Returns the name (not full identifier) of the object
        being documented

        :return:
        :rtype:
        """
        return self.get_name()

    def get_name(self):
        """
        Returns the name the object will have in its documentation page

        :return:
        :rtype:
        """
        if self._name is not None:
            name = self._name
        else:
            try:
                name = self.obj.__name__
            except AttributeError:
                name = "<{} Instance>".format(type(self.obj).__name__)

        return name

    @classmethod
    def get_identifier(cls, o):
        try:
            pkg = o.__module__
        except AttributeError:
            pkg = ""

        try:
            n = o.__qualname__
        except AttributeError:
            try:
                n = o.__name__
            except AttributeError:
                n = type(o).__name__

        qn = pkg + ('.' if pkg != "" else "") + n
        return qn
    @property
    def identifier(self):
        if self._id is None:
            self._id = self.get_identifier(self.obj)
        return self._id

    @property
    def parent(self):
        """
        Returns the parent object for docs purposes

        :return:
        :rtype:
        """
        if self._pobj is None:
            self._pobj = self.resolve_parent()
        return self._pobj
    def resolve_parent(self, check_tree=True):
        """
        Resolves the "parent" of obj.
        By default, just the module in which it is contained.
        Allows for easy skipping of pieces of the object tree,
        though, since a parent can be directly added to the set of
        written object which is distinct from the module it would
        usually resolve to.
        Also can be subclassed to provide more fine grained behavior.

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        if check_tree:
            oid = self.identifier
            if self.tree is not None and oid in self.tree:
                return self.tree[oid]['parent']

        if self._parent is not None:
            if isinstance(self._parent, str):
                return self.walker.resolve_object(self._parent)
            else:
                return self._parent
        elif 'parent' in self.spec:
            parent = self.spec['parent']
            if isinstance(parent, str):
                return self.walker.resolve_object(parent)
            else:
                return parent

        if isinstance(self.obj, types.ModuleType):
            # not totally sure this will work...
            modspec = self.obj.__name__.rsplit(".", 1)[0]
        else:
            try:
                modspec = self.obj.__module__
            except AttributeError:
                modspec = type(self.obj).__module__

        if modspec == "":
            return None

        return self.walker.resolve_object(modspec)

    def resolve_relative_obj(self, spec:str):

        if '.' == spec[0]:
            n = 0
            while spec[0] == ".":
                n += 1
                spec = spec[1:]
            if isinstance(self.obj, types.ModuleType):
                # not totally sure this will work...
                modspec = self.obj.__name__
            else:
                try:
                    modspec = self.obj.__module__
                except AttributeError:
                    modspec = type(self.obj).__module__
            modspec = modspec.rsplit(".", n)[0]
            o = self.walker.resolve_object(modspec+"."+spec)
        else:
            bits = spec.split(".")
            try:
                o = getattr(self.obj, bits[0])
            except AttributeError:
                try:
                    modspec = self.obj.__module__
                except AttributeError:
                    modspec = type(self.obj).__module__
                o = getattr(self.walker.resolve_object(modspec), bits[0])
            for s in bits[1:]:
                o = getattr(o, s)
        return o

    @property
    def children(self):
        """
        Returns the child objects for docs purposes

        :return:
        :rtype:
        """
        if self._chobj is None:
            self._chobj = self.resolve_children()
        return self._chobj
    def resolve_children(self, check_tree=True):
        """
        Resolves the "children" of obj.
        First tries to use any info supplied by the docs tree
        or a passed object spec, then that failing looks for an
        `__all__` attribute

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        childs = None
        if check_tree:
            oid = self.identifier
            if self.tree is not None and oid in self.tree:
                childs = self.tree[oid]['children']
        if childs is None:
            if 'children' in self.spec:
                childs = self.spec['children']
            elif hasattr(self.obj, '__all__'):
                childs = self.obj.__all__
            else:
                childs = []

        oid = self.identifier
        return [self.walker.resolve_object(oid + "." + x) if isinstance(x, str) else x for x in childs]

    @property
    def tree_spec(self):
        """
        Provides info that gets added to the `written` dict and which allows
        for a doc tree to be built out.

        :return:
        :rtype:
        """

        base_spec = {
            'id': self.identifier,
            'parent': self.parent,
            'children': self.children
        }
        base_spec.update(self.spec)
        return base_spec

    @abc.abstractmethod
    def handle(self):
        raise NotImplementedError("abstract method")
    def stop_traversal(self):
        return False

class ObjectWalker:
    """
    A class that walks a module/object structure, calling handlers
    appropriately at each step

    A class that walks a module structure, generating .md files for every class inside it as well as for global functions,
    and a Markdown index file.

    Takes a set of objects & writers and walks through the objects, generating files on the way
    """

    spec = ObjectSpec
    default_handlers = collections.OrderedDict()
    def __init__(self,
                 handlers=None,
                 tree=None,
                 **extra_fields
                 ):
        """
        :param objects: the objects to write out
        :type objects: Iterable[Any]
        :param out: the directory in which to write the files (`None` means `sys.stdout`)
        :type out: None | str
        :param out: the directory in which to write the files (`None` means `sys.stdout`)
        :type out: None | str
        :param: writers
        :type: DispatchTable
        :param ignore_paths: a set of paths not to write (passed to the objects)
        :type ignore_paths: None | Iterable[str]
        """

        # obtain default writer set
        if handlers is None:
            handlers = {}
        if not isinstance(handlers, MethodDispatch):
            fallback_handler = ObjectHandler if None not in self.default_handlers else self.default_handlers[None]
            if hasattr(handlers, 'items'):
                handlers = MethodDispatch(handlers.items(), default=fallback_handler)
            else:
                handlers = MethodDispatch(handlers, default=fallback_handler)
        for k, v in self._initial_handlers.items():
            if k not in handlers:
                handlers[k] = v
        self.handlers = handlers

        # obtain default tree
        if tree is None:
            tree = ObjectTree()
        self.tree = tree

        self.extra_fields = extra_fields

    @property
    def _initial_handlers(self):
        """
        Adds a minor hook onto the default_writes dict and returns it
        :return:
        :rtype:
        """

        writers = self.default_handlers.copy()
        writers[self.spec] = self.resolve_spec
        return writers

    def get_handler(self, obj, *, tree=None, walker=None, cls=None, **kwargs):
        if cls is not None:
            return cls(obj, **dict(self.extra_fields, walker=self, tree=self.tree, **kwargs))
        else:
            return self.handlers(obj,
                                 **dict(self.extra_fields, walker=self, tree=self.tree, **kwargs)
                                 )
    @staticmethod
    def resolve_object(o):
        """
        Resolves to an arbitrary object by name

        :param o:
        :type o:
        :return:
        :rtype:
        """

        if o in sys.modules:
            # first we try to do a direct look up
            o = sys.modules[o]
        elif o.rsplit(".", 1)[0] in sys.modules:
            # if direct lookup failed, but the parent module has been loaded
            # try direct lookup on that
            try:
                mod, attr = o.rsplit(".", 1)
                o = getattr(sys.modules[mod], attr)
            except AttributeError:
                o = importlib.import_module(o)
        else:
            # otherwise fall back on standard import machinery
            try:
                o = importlib.import_module(o)
            except ModuleNotFoundError:  # tried to load a member but couldn't...
                # we try to resolve this first by importing as deep as we can and then
                # requesting the final attribute pieces
                p_split = o.split(".")
                for i in range(1, len(p_split)):
                    mod_spec = ".".join(p_split[:-i])
                    try:
                        mood = importlib.import_module(mod_spec)
                    except ModuleNotFoundError:
                        pass
                    else:
                        try:
                            from functools import reduce
                            v = reduce(lambda m, a: getattr(m, a), p_split[-i:], mood)
                        except AttributeError:
                            pass
                        else:
                            o = v
                            break
                else:
                    raise ValueError("couldn't resolve {}".format(o))
        return o

    def resolve_spec(self, spec, **kwargs):
        """
        Resolves an object spec.

        :param spec: object spec
        :type spec: ObjectSpec
        :return:
        :rtype:
        """

        # for the moment we only reolve using the `id` parameter
        oid = spec['id']
        o = self.resolve_object(oid)
        return self.get_handler(o, spec=spec, **kwargs)

    def visit(self, o, parent=None, depth=0, max_depth=-1, **kwargs): # TODO: add traversal control
        """
        Visits a single object in the tree
        Provides type dispatching to a handler, basically.

        :param o: the object we want to handler
        :type o: Any
        :param parent: the handler that was called right before this
        :type parent: ObjectHandler
        :return: the result of handling
        :rtype: Any
        """

        if max_depth < 0 or depth < max_depth:

            if (
                    isinstance(o, (dict, collections.OrderedDict))
                    and all(k in o for k in self.spec.required_keys)
            ):
                o = self.spec(o.items())

            if parent is not None:
                pid = parent.identifier
            else:
                pid = None

            if isinstance(o, self.spec):
                handler = self.resolve_spec(o, parent=pid, **kwargs)
            else:
                handler = self.get_handler(o, parent=pid, **kwargs)
            oid = handler.identifier
            if oid not in self.tree:
                if handler.stop_traversal():
                    res = None
                else:
                    spec = handler.tree_spec
                    self.tree[oid] = spec

                    for child in handler.children: # depth first traversal
                        self.visit(child, parent=handler, depth=depth+1, max_depth=max_depth, **dict(handler.extra_fields, **kwargs))

                    res = handler.handle()
                    spec.update(output=res)

                return res


import enum

__all__ = [
    "Options"
]

class Options(enum.Enum):
    BackgroundAttachment = "background-attachment"
    BackgroundBlendMode = "background-blend-mode"
    BackgroundColor = "background-color"
    BackgroundImage = "background-image"
    BackgroundPosition = "background-position"
    BackgroundRepeat = "background-repeat"
    Border = "border"
    BorderBottom = "border-bottom"
    BorderBottomColor = "border-bottom-color"
    BorderBottomStyle = "border-bottom-style"
    BorderBottomWidth = "border-bottom-width"
    BorderCollapse = "border-collapse"
    BorderColor = "border-color"
    BorderLeft = "border-left"
    BorderLeftColor = "border-left-color"
    BorderLeftStyle = "border-left-style"
    BorderLeftWidth = "border-left-width"
    BorderRight = "border-right"
    BorderRightColor = "border-right-color"
    BorderRightStyle = "border-right-style"
    BorderRightWidth = "border-right-width"
    BorderSpacing = "border-spacing"
    BorderStyle = "border-style"
    BorderTop = "border-top"
    BorderTopColor = "border-top-color"
    BorderTopStyle = "border-top-style"
    BorderTopWidth = "border-top-width"
    BorderWidth = "border-width"
    Bottom = "bottom"
    BreakAfter = "break-after"
    BreakBefore = "break-before"
    BreakInside = "break-inside"
    CaptionSide = "caption-side"
    Clear = "clear"
    Clip = "clip"
    Color = "color"
    Content = "content"
    CounterIncrement = "counter-increment"
    CounterReset = "counter-reset"
    Cursor = "cursor"
    Direction = "direction"
    Display = "display"
    EmptyCells = "empty-cells"
    Float = "float"
    Font = "font"
    FontFamily = "font-family"
    FontSize = "font-size"
    FontStyle = "font-style"
    FontVariant = "font-variant"
    FontWeight = "font-weight"
    Grid = "grid"
    GridArea = "grid-area"
    GridAutoColumns = "grid-auto-columns"
    GridAutoFlow = "grid-auto-flow"
    GridAutoRows = "grid-auto-rows"
    GridColumn = "grid-column"
    GridColumnEnd = "grid-column-end"
    GridColumnGap = "grid-column-gap"
    GridColumnStart = "grid-column-start"
    GridGap = "grid-gap"
    GridRow = "grid-row"
    GridRowEnd = "grid-row-end"
    GridRowGap = "grid-row-gap"
    GridRowStart = "grid-row-start"
    GridTemplate = "grid-template"
    GridTemplateAreas = "grid-template-areas"
    GridTemplateColumns = "grid-template-columns"
    GridTemplateRows = "grid-template-rows"
    Height = "height"
    Left = "left"
    LetterSpacing = "letter-spacing"
    LineHeight = "line-height"
    ListStyle = "list-style"
    ListStyleImage = "list-style-image"
    ListStylePosition = "list-style-position"
    ListStyleType = "list-style-type"
    Margin = "margin"
    MarginBottom = "margin-bottom"
    MarginLeft = "margin-left"
    MarginRight = "margin-right"
    MarginTop = "margin-top"
    MaxHeight = "max-height"
    MaxWidth = "max-width"
    MinHeight = "min-height"
    MinWidth = "min-width"
    MixBlendMode = "mix-blend-mode"
    Outline = "outline"
    OutlineColor = "outline-color"
    OutlineStyle = "outline-style"
    OutlineWidth = "outline-width"
    Overflow = "overflow"
    Padding = "padding"
    PaddingBottom = "padding-bottom"
    PaddingLeft = "padding-left"
    PaddingRight = "padding-right"
    PaddingTop = "padding-top"
    PageBreakAfter = "page-break-after"
    PageBreakBefore = "page-break-before"
    PageBreakInside = "page-break-inside"
    Position = "position"
    Quotes = "quotes"
    Right = "right"
    ScrollBehavior = "scroll-behavior"
    TableLayout = "table-layout"
    TextAlign = "text-align"
    TextDecoration = "text-decoration"
    TextDecorationThickness = "text-decoration-thickness"
    TextIndent = "text-indent"
    TextTransform = "text-transform"
    Top = "top"
    UnicodeBidi = "unicode-bidi"
    VerticalAlign = "vertical-align"
    Visibility = "visibility"
    WhiteSpace = "white-space"
    Width = "width"
    WordSpacing = "word-spacing"
    ZIndex = "z-index"



from nbconvert.preprocessors.extractoutput import *

class MarkdownImageExtractor(ExtractOutputPreprocessor):
    """
    Extracts all of the outputs from the notebook file.  The extracted
    outputs are returned in the 'resources' dictionary.
    """

    def __init__(self, *args, prefix='', **kwargs):
        super().__init__(*args, **kwargs)
        self.prefix = prefix

    def reencode_data(self, mime_type, data):
        # Binary files are base64-encoded, SVG is already XML
        if mime_type in {'image/png', 'image/jpeg', 'application/pdf'}:
            # data is b64-encoded as text (str, unicode),
            # we want the original bytes
            data = a2b_base64(data)
        elif mime_type == 'application/json' or not isinstance(data, text_type):
            # Data is either JSON-like and was parsed into a Python
            # object according to the spec, or data is for sure
            # JSON. In the latter case we want to go extra sure that
            # we enclose a scalar string value into extra quotes by
            # serializing it properly.
            if isinstance(data, bytes) and not isinstance(data, text_type):
                # In python 3 we need to guess the encoding in this
                # instance. Some modules that return raw data like
                # svg can leave the data in byte form instead of str
                data = data.decode('utf-8')
            data = platform_utf_8_encode(json.dumps(data))
        else:
            # All other text_type data will fall into this path
            data = platform_utf_8_encode(data)
        return data

    def save_attachement(self, filename, attachement, index, cell, resources, cell_index):

        # Get the unique key from the resource dict if it exists.  If it does not
        # exist, use 'output' as the default.  Also, get files directory if it
        # has been specified
        unique_key = resources.get('unique_key', 'output')
        output_files_dir = resources.get('output_files_dir', None)

        mime_type = next(iter(attachement.keys()))
        data = next(iter(attachement.values()))

        # ext = guess_extension_without_jpe(mime_type)
        # if ext is None:
        #     ext = '.' + mime_type.rsplit('/')[-1]
        # if out.metadata.get('filename', ''):
        #     filename = out.metadata['filename']
        #     if not filename.endswith(ext):
        #         filename += ext
        # else:
        # filename = self.output_filename_template.format(
        #     unique_key=unique_key,
        #     cell_index=cell_index,
        #     index=index,
        #     extension=ext)

        resources['outputs'][filename] = self.reencode_data(mime_type, data)
        return filename

    def preprocess_cell(self, cell, resources, cell_index):
        if cell['cell_type'] == 'markdown':
            if 'attachments' in cell:
                for index, (name, attachement) in enumerate(cell['attachments'].items()):
                    fn = self.save_attachement(name, attachement, index, cell, resources, cell_index)
                    cell['source'] = cell['source'].replace(
                        'attachment:'+name,
                        self.prefix+fn
                    )
        elif cell['cell_type'] == 'code' and cell['source'].strip() == "":
            cell['cell_type'] = 'raw'
            cell['source'] = cell['source'].strip()
        else:
            cell, resources = super().preprocess_cell(cell, resources, cell_index)

        return cell, resources




from .VectorOps import vec_normalize, vec_angles
import math, numpy as np, scipy as sp

__all__ = [
    "rotation_matrix",
    "skew_symmetric_matrix",
    "rotation_matrix_skew",
    "translation_matrix",
    "affine_matrix",
    "extract_rotation_angle_axis"
]

#######################################################################################################################
#
#                                                 rotation_matrix
#

def rotation_matrix_basic(xyz, theta):
    """rotation matrix about x, y, or z axis

    :param xyz: x, y, or z axis
    :type xyz: str
    :param theta: counter clockwise angle in radians
    :type theta: float
    """

    axis = xyz.lower()
    if axis == "z": # most common case so it comes first
        mat = [
            [ math.cos(theta), -math.sin(theta), 0.],
            [ math.sin(theta),  math.cos(theta), 0.],
            [ 0.,               0.,              1.]
        ]
    elif axis == "y":
        mat = [
            [ math.cos(theta), 0., -math.sin(theta)],
            [ 0.,              1.,               0.],
            [ math.sin(theta), 0.,  math.cos(theta)]
        ]
    elif axis == "x":
        mat = [
            [ 1.,               0.,               0.],
            [ 0.,  math.cos(theta), -math.sin(theta)],
            [ 0.,  math.sin(theta),  math.cos(theta)]
        ]
    else:
        raise Exception("{}: axis '{}' invalid".format('rotation_matrix_basic', xyz))
    return np.array(mat)

def rotation_matrix_basic_vec(xyz, thetas):
    """rotation matrix about x, y, or z axis

    :param xyz: x, y, or z axis
    :type xyz: str
    :param thetas: counter clockwise angle in radians
    :type thetas: float
    """

    thetas = np.asarray(thetas)
    nmats = len(thetas)
    z = np.zeros((nmats,))
    o = np.ones((nmats,))
    c = np.cos(thetas)
    s = np.sin(thetas)
    axis = xyz.lower()
    if axis == "z": # most common case so it comes first
        mat = [
            [ c, s, z],
            [-s, c, z],
            [ z, z, o]
        ]
    elif axis == "y":
        mat = [
            [ c, z, s],
            [ z, o, z],
            [-s, z, c]
        ]
    elif axis == "x":
        mat = [
            [ o, z, z],
            [ z, c, s],
            [ z,-s, c]
        ]
    else:
        raise Exception("{}: axis '{}' invalid".format('rotation_matrix_basic', xyz))
    return np.array(mat).T

#thank you SE for the nice Euler-Rodrigues imp: https://stackoverflow.com/questions/6802577/rotation-of-3d-vector
def rotation_matrix_ER(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    """
    axis = np.asarray(axis)
    axis = axis / math.sqrt(np.dot(axis, axis))
    a = math.cos(theta / 2.0)
    b, c, d = -axis * math.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([
        [aa + bb - cc - dd, 2 * (bc + ad),     2 * (bd - ac)    ],
        [2 * (bc - ad),     aa + cc - bb - dd, 2 * (cd + ab)    ],
        [2 * (bd + ac),     2 * (cd - ab),     aa + dd - bb - cc]
    ])

def rotation_matrix_ER_vec(axes, thetas):
    """
    Vectorized version of basic ER
    """

    axes = np.asanyarray(axes)
    thetas = np.asanyarray(thetas)
    # if len(axes.shape) == 1:
    #     axes = axes/np.linalg.norm(axes)
    #     axes = np.broadcast_to(axes, (len(thetas), 3))
    # else:
    #     axes = vec_normalize(axes)

    ax_shape = axes.shape[:-1]
    t_shape = thetas.shape
    axes = np.reshape(axes, (-1, 3))
    thetas = thetas.reshape(-1)
    if thetas.ndim == 0:
        base_shape = ax_shape
    elif axes.ndim == 1:
        base_shape = t_shape
    elif thetas.ndim != axes.ndim - 1:
        raise ValueError(f"can't broadcast axes and angles with shapes {ax_shape} and {t_shape}")
    else:
        base_shape = tuple(a if t == 1 else t for a,t in zip(ax_shape, t_shape))

    a = np.cos(thetas/2.0)
    b, c, d = np.moveaxis(-axes * np.reshape(np.sin(thetas / 2.0), (len(thetas), 1)), -1, 0)
    v = np.array([a, b, c, d])
    # triu_indices
    rows, cols = (
        np.array([0, 0, 0, 0, 1, 1, 1, 2, 2, 3]),
        np.array([0, 1, 2, 3, 1, 2, 3, 2, 3, 3])
    )
    aa, ab, ac, ad, bb, bc, bd, cc, cd, dd = v[rows] * v[cols]
    ## Uses half-angle formula to get compact form for Euler-Rodrigues
    # a^2 * I + [[ 0,    2ad, -2ac]   + [[b^2 - c^2 - d^2,               2bc,              2bd]
    #            [-2ad,    0,  2ab],     [             2bc, -b^2 + c^2 - d^2,              2cd]
    #            [ 2ac, -2ab,    0]]     [             2bd,              2cd, -b^2 - c^2 + d^2]]
    R = np.array([
        [aa + bb - cc - dd,      2 * (bc + ad),         2 * (bd - ac)],
        [    2 * (bc - ad),  aa - bb + cc - dd,         2 * (cd + ab)],
        [    2 * (bd + ac),      2 * (cd - ab),     aa - bb - cc + dd]
    ])
    R = np.moveaxis(R, -1, 0)

    return R.reshape(base_shape + (3, 3))

def rotation_matrix_align_vectors(vec1, vec2):
    angles, normals = vec_angles(vec1, vec2)
    return rotation_matrix(normals, angles)

def rotation_matrix(axis, theta):
    """
    :param axis:
    :type axis:
    :param theta: angle to rotate by (or Euler angles)
    :type theta:
    :return:
    :rtype:
    """

    try:
        flen = len(theta)
    except TypeError:
        flen = 0
    extra_shape = None

    if type(axis) == str:
        if flen >0:
            mat_fun = rotation_matrix_basic_vec
        else:
            mat_fun = rotation_matrix_basic
    else:
        if flen > 0:
            axis = np.asanyarray(axis)
            theta = np.asanyarray(theta)

            if axis.ndim == theta.ndim:
                if theta.ndim > 2:
                    extra_shape = theta.shape[:-2]
                    axis = axis.reshape((-1, 3))
                    theta = theta.reshape((-1, 3))
                mat_fun = rotation_matrix_align_vectors
            else:
                if theta.ndim > 1:
                    extra_shape = theta.shape
                    axis = axis.reshape((-1, 3))
                    theta = theta.reshape(-1)
                mat_fun = rotation_matrix_ER_vec
        else:
            mat_fun = rotation_matrix_ER

    mats = mat_fun(axis, theta)
    if extra_shape is not None:
        mats = mats.reshape(extra_shape + (3, 3))
    return mats

def skew_symmetric_matrix(upper_tri):
    upper_tri = np.asanyarray(upper_tri)
    l = len(upper_tri)
    n = int((1 + np.sqrt(1 + 8*l)) // 2)
    m = np.zeros((n, n))
    rows, cols = np.triu_indices_from(m, 1)
    m[rows, cols] =  upper_tri
    m[cols, rows] = -upper_tri
    return m

def extract_rotation_angle_axis(rot_mat):
    ang = np.arccos((np.trace(rot_mat) - 1)/2)
    skew = (rot_mat - rot_mat.T)/2
    ax = np.array([skew[2, 1], skew[0, 2], skew[1, 0]])
    n = np.linalg.norm(ax)
    return ang, ax/n

def youla_skew_decomp(A):
    n = len(A)
    s, T = sp.linalg.schur(A)

    l = np.diag(s, 1)
    if n % 2 == 0:
        start = 0
        end = n
    else:  # manage padding for odd dimension
        if abs(l[0]) < 1e-15:
            start = 1
            end = n
        else:
            start = 0
            end = n - 1
        l = l[start:end-1]
    cos = np.cos(l)
    sin = np.sin(l)

    # build 2x2 block mat
    U = np.eye(n)
    o = np.arange(start, end, 2)  # even inds
    e = np.arange(start + 1, end, 2)  # odd inds

    # print(start, end, l)

    U[o, o] = cos
    U[e, e] = cos
    U[o, e] = sin
    U[e, o] = -sin

    return U, T

def rotation_matrix_skew(upper_tri):
    A = skew_symmetric_matrix(upper_tri)
    # build Youla matrix
    U, T = youla_skew_decomp(A)
    return T@U@T.T

#######################################################################################################################
#
#                                                 translation_matrix
#

def translation_matrix(shift):
    share = np.asarray(shift)
    if len(share.shape) == 1:
        ss = share
        zs = 0.
        os = 1.
        mat = np.array(
            [
                [os, zs, zs, ss[0]],
                [zs, os, zs, ss[1]],
                [zs, zs, os, ss[2]],
                [zs, zs, zs, os   ]
            ]
        )
    else:
        zs = np.zeros((share.shape[0],))
        os = np.ones((share.shape[0],))
        ss = share.T
        mat = np.array(
            [
                [os, zs, zs, ss[0]],
                [zs, os, zs, ss[1]],
                [zs, zs, os, ss[2]],
                [zs, zs, zs, os   ]
            ]
        ).T
    return mat

#######################################################################################################################
#
#                                                 affine_matrix
#

def affine_matrix(tmat, shift):
    """Creates an affine transformation matrix from a 3x3 transformation matrix or set of matrices and a shift or set of vecs

    :param tmat: base transformation matrices
    :type tmat: np.ndarray
    :param shift:
    :type shift:
    :return:
    :rtype:
    """

    base_mat = np.asanyarray(tmat)
    if shift is None:
        return base_mat

    if base_mat.ndim > 2:
        shifts = np.asanyarray(shift)
        if shifts.ndim == 1:
            shifts = np.broadcast_to(shifts, (1,)*(base_mat.ndim-2) + shifts.shape)
        shifts = np.broadcast_to(shifts, base_mat.shape[:-2] + (3,))
        shifts = np.expand_dims(shifts, -1)
        mat = np.concatenate([base_mat, shifts], axis=-1)
        padding = np.array([0., 0., 0., 1.])
        padding = np.broadcast_to(
            np.broadcast_to(padding, (1,)*(base_mat.ndim-2) + padding.shape),
            mat.shape[:-2] + (4,)
        )
        padding = np.expand_dims(padding, -2)
        mat = np.concatenate([mat, padding], axis=-2)
    else:
        mat = np.concatenate([base_mat, shift[:, np.newaxis]], axis=-1)
        mat = np.concatenate([mat, [[0., 0., 0., 1.]]], axis=-2)
    return mat


from abc import ABCMeta, abstractmethod

######################################################################################################
##
##                                   TransformationFunction Class
##
######################################################################################################

__all__ = [
    "TransformationFunction"
]

class TransformationFunction(metaclass=ABCMeta):
    """
    The TransformationFunction class is an abstract class
    It provides the scaffolding for representing a single transformation operation
    """

    def __init__(self):
        """Initializes a transformation function based on the transfdata

        :param transfdata:
        :type transfdata:
        """
        pass

    @property
    def inverse(self):
        """
        Returns the inverse of the transformation
        :return:
        :rtype:
        """
        raise NotImplementedError("arbitrary transform inverses not implemented")

    @abstractmethod
    def merge(self, other):
        """Tries to merge with another TransformationFunction

        :param other: a TransformationFunction to try to merge with
        :type other: TransformationFunction
        :return: tfunc
        :rtype: TransformationFunction
        """
        pass

    @abstractmethod
    def operate(self, coords, shift=True):
        """Operates on the coords. *Must* be able to deal with a list of coordinates, optimally in an efficient manner

        :param coords: the list of coordinates to apply the transformation to
        :type coords: np.ndarry
        """
        pass



from .HTML import HTML, CSS
from .BootstrapEnums import *

__all__ = ["Bootstrap"]
__reload_hook__ = [".HTML"]

class BootstrapBase(HTML):
    # Bootstrap components
    @classmethod
    def _add_cls(kls, base_cls, cls):
        if base_cls is None:
            base_cls = cls
        else:
            if isinstance(base_cls, str):
                base_cls = base_cls.split()
            try:
                base_cls = list(base_cls)
            except TypeError:
                base_cls = [str(base_cls)]
            if isinstance(cls, str):
                cls = cls.split()
            try:
                cls = list(cls)
            except TypeError:
                cls = [str(cls)]
            for c in cls:
                if c not in base_cls:
                    base_cls.append(c)
        return base_cls
    @classmethod
    def _manage_cls(kls, obj, cls, variant):
        added_cls = obj.cls
        if variant is None:
            variant = obj.base_style
        if variant is not None:
            if isinstance(variant, (str, BootstrapBase.Variant)):
                variant = [variant]
            added_cls = [obj.cls, *(obj.cls + "-" + str(v) for v in variant)]
        return BootstrapBase._add_cls(cls, added_cls)
    class DivComponent(HTML.Div):
        cls = None
        base_style = None
        def __init__(self, *elems, variant=None, cls=None, **attrs):
            cls = BootstrapBase._manage_cls(self, cls, variant)
            super().__init__(*elems, cls=cls, **attrs)
    class SpanComponent(HTML.Span):
        cls = None
        base_style = None
        def __init__(self, *elems, variant=None, cls=None, **attrs):
            cls = BootstrapBase._manage_cls(self, cls, variant)
            super().__init__(*elems, cls=cls, **attrs)
    class ListComponent(HTML.List):
        cls = None
        base_style = None
        def __init__(self, *elems, variant=None, cls=None, **attrs):
            cls = BootstrapBase._manage_cls(self, cls, variant)
            super().__init__(*elems, cls=cls, **attrs)
    class ListItemComponent(HTML.ListItem):
        cls = None
        base_style = None
        def __init__(self, *elems, variant=None, cls=None, **attrs):
            cls = BootstrapBase._manage_cls(self, cls, variant)
            super().__init__(*elems, cls=cls, **attrs)
    class ItalicComponent(HTML.Italic):
        cls = None
        base_style = None
        def __init__(self, *elems, variant=None, cls=None, **attrs):
            cls = BootstrapBase._manage_cls(self, cls, variant)
            super().__init__(*elems, cls=cls, **attrs)
    class TableComponent(HTML.Table):
        cls = None
        def __init__(self, *elems, variant=None, cls=None, **attrs):
            cls = BootstrapBase._manage_cls(self, cls, variant)
            super().__init__(*elems, cls=cls, **attrs)
    class ButtonComponent(HTML.Button):
        cls = None
        def __init__(self, *elems, variant=None, cls=None, **attrs):
            cls = BootstrapBase._manage_cls(self, cls, variant)
            super().__init__(*elems, cls=cls, **attrs)
    class AnchorComponent(HTML.Anchor):
        cls = None
        def __init__(self, *elems, variant=None, cls=None, **attrs):
            cls = BootstrapBase._manage_cls(self, cls, variant)
            super().__init__(*elems, cls=cls, **attrs)

    class BoostrapIcon(ItalicComponent):
        cls='bi'
        def __init__(self, icon_name, cls=None, **attrs):
            # cls = BootstrapBase._manage_cls(self, cls, icon_name)
            super().__init__(" ", variant=icon_name, cls=cls, **attrs)
    class FontAwesomeIcon(SpanComponent):
        cls='fa'
        def __init__(self, icon_name, **attrs):
            super().__init__(" ", variant=icon_name, **attrs)
    class GlyphIcon(SpanComponent):
        cls='glyphicon'
        def __init__(self, icon_name, **attrs):
            super().__init__(" ", variant=icon_name, **attrs)

    class Col(HTML.Div):
        cls='col'
        base_size=None
        base_width=None
        def __init__(self, *elems, width=None, size=None, cls=None, **attrs):
            added_cls = self.cls
            if size is None:
                size = self.base_size
            if size is not None:
                added_cls += '-'+str(size)
            if width is None:
                width = self.base_width
            if width is not None:
                added_cls += '-'+str(width)
            cls = BootstrapBase._add_cls(cls, added_cls)
            super().__init__(*elems, cls=cls, **attrs)
        def __repr__(self):
            return "{}({})".format(type(self).__name__, self.elems)
    class Row(DivComponent):
        cls = 'row'
        def __init__(self, *cols, item_attributes=None, **attrs):
            if len(cols) == 1 and isinstance(cols[0], (list, tuple)):
                cols = cols[0]
            if item_attributes is None:
                item_attributes = {}
            super().__init__(
                [BootstrapBase.Col(x, **item_attributes) if not isinstance(x, BootstrapBase.Col) else x for x in cols],
                **attrs
            )
    class Container(DivComponent): cls='container'
    @staticmethod
    def Grid(rows, row_attributes=None, item_attributes=None, auto_size=False, **attrs):
        if row_attributes is None:
            row_attributes = {}
        if item_attributes is None:
            item_attributes = {}
        if auto_size:
            # most naive approach...
            width = round(100/max(len(r) for r in rows))
            item_attributes['style'] = CSS(width=str(width)+'%')
            attrs['style'] = CSS(width='100%')
        return BootstrapBase.Container(
            [BootstrapBase.Row(r, item_attributes=item_attributes, **row_attributes) for r in rows],
            **attrs
        )

    class Alert(DivComponent): cls='alert'
    class Breadcrumb(ListComponent): cls = ['breadcrumb']
    class BreadcrumbItem(ListItemComponent): cls = ['breadcrumb-item']
    class ListGroup(ListComponent): cls='list-group'
    class ListGroupItem(ListItemComponent): cls='list-group-item'

    class ButtonGroup(DivComponent):
        cls = 'btn-group'
    class Button(ButtonComponent):
        cls = 'btn'
        base_style = 'primary'
    class CloseButton(ButtonComponent):
        cls = 'btn-close'
    class LinkButton(AnchorComponent):
        cls = 'btn'
        base_style = 'default'
    class Table(TableComponent):
        cls = 'table'
        base_style = 'hover'

    class Label(SpanComponent):
        cls = 'label'
        base_style = 'default'
    class Badge(SpanComponent): cls = 'badge'

    Class = SemanticClass
    Variant = SemanticVariant

class Bootstrap3(BootstrapBase):
    Icon = BootstrapBase.GlyphIcon

    class Jumbotron(BootstrapBase.DivComponent): cls='jumbotron'

    class PanelBody(BootstrapBase.DivComponent): cls='panel-body'
    class PanelHeader(BootstrapBase.DivComponent): cls='panel-heading'
    class Panel(BootstrapBase.DivComponent):
        cls='panel'
        base_style='default'
        def __init__(self, *elems, header=None, **attrs):
            if header is None:
                super().__init__(*elems, **attrs)
            else:
                header = Bootstrap3.PanelHeader(header)
                super().__init__(header, Bootstrap3.PanelBody(*elems), **attrs)

class Bootstrap4(BootstrapBase):
    Icon = BootstrapBase.FontAwesomeIcon

    class Jumbotron(BootstrapBase.DivComponent): cls='jumbotron'

    class Pill(BootstrapBase.SpanComponent): cls = ['badge', 'badge-pill']

    class Collapse(BootstrapBase.DivComponent): cls='collapse'

    class CardBody(BootstrapBase.DivComponent): cls='card-body'
    class CardHeader(BootstrapBase.DivComponent): cls='card-header'
    class CardFooter(BootstrapBase.DivComponent): cls='card-footer'
    class CardImage(BootstrapBase.DivComponent): cls='card-img-top'
    class Card(BootstrapBase.DivComponent):
        cls='card'
        base_style=None
        def __init__(self, *elems, header=None, **attrs):
            if header is None:
                super().__init__(*elems, **attrs)
            else:
                header = Bootstrap4.CardHeader(header)
                super().__init__(header, Bootstrap4.CardBody(*elems), **attrs)

class Bootstrap5(BootstrapBase):
    Icon = BootstrapBase.BoostrapIcon

    class Pill(BootstrapBase.SpanComponent): cls = ['badge', 'rounded-pill']

    class Accordion(BootstrapBase.DivComponent): cls='accordion'
    class AccordionItem(BootstrapBase.DivComponent): cls='accordion-item'
    class AccordionCollapse(BootstrapBase.DivComponent): cls='accordion-collapse'
    class AccordionBody(BootstrapBase.DivComponent): cls='accordion-body'

    class Carousel(BootstrapBase.DivComponent): cls='carousel'
    class CarouselInner(BootstrapBase.DivComponent): cls='carousel-inner'
    class CarouselItem(BootstrapBase.DivComponent): cls='carousel-item'

    class Collapse(BootstrapBase.DivComponent): cls='collapse'

    class CardBody(BootstrapBase.DivComponent): cls='card-body'
    class CardHeader(BootstrapBase.DivComponent): cls='card-header'
    class CardFooter(BootstrapBase.DivComponent): cls='card-footer'
    class CardImage(BootstrapBase.DivComponent): cls='card-img-top'
    class Card(BootstrapBase.DivComponent):
        cls='card'
        base_style=None
        def __init__(self, *elems, header=None, **attrs):
            if header is None:
                super().__init__(*elems, **attrs)
            else:
                header = Bootstrap5.CardHeader(header)
                super().__init__(header, Bootstrap5.CardBody(*elems), **attrs)

Bootstrap = Bootstrap5


"""
A module for making use of the results of calculations run by the Gaussian electronic structure package.
We'd like to be able to also support the NWChem and Psi4 packages, but haven't had the time, yet, to write it out.

Two main avenues of support are provided:

    1. importing Gaussian results
    2. setting up Gaussian jobs

The first is likely to be more useful to you, but we're hoping to be able to hook (2.) into the `Psience.Molecools` package.
The goal there is to provide automated support for setting up scans of molecular vibrations & the like.

There are already direct hooks into (1.) in `Psience.Data` through the `DipoleSurface` and `PotentialSurface` objects.
These are still in the prototype stage, but hopefully will allow us to unify strands of our Gaussian support,
 and also make it easy to unify support for Psi4 and NWChem data, once we have the basic interface down.
"""
from .GaussianImporter import *
from .GaussianJob import *
from .FChkDerivatives import *

__all__ = []
from .GaussianImporter import __all__ as exposed
__all__ += exposed
from .GaussianJob import __all__ as exposed
__all__ += exposed
from .FChkDerivatives import __all__ as exposed
__all__ += exposed


"""
Just a simple text table formatter with support for headers, separators, any kind of python formatting spec
etc.
"""

import numpy as np

__all__ = [
    "TableFormatter"
]

class TableFormatter:

    default_header_format=""
    default_column_join = '  '
    default_row_join = '\n'
    default_separator = "-"
    default_column_alignment = '.'
    default_header_alignment = '^'
    default_row_padding = ""
    def __init__(self,
                 column_formats,
                 *,
                 headers=None,
                 header_spans=None,
                 header_format=None,
                 column_join=None,
                 row_join=None,
                 header_column_join=None,
                 header_row_join=None,
                 separator=None,
                 separator_lines=1,
                 content_join=None,
                 column_alignments=None,
                 header_alignments=None,
                 row_padding=None
                 ):
        self.headers = headers
        self.header_spans = header_spans
        self.header_format = header_format
        self.column_formats = column_formats
        self.column_alignments = column_alignments
        self.header_alignments = header_alignments
        self.column_join = column_join
        self.row_join = row_join
        self.header_column_join = header_column_join
        self.header_row_join = header_row_join
        self.separator = separator
        self.separator_lines = separator_lines
        self.content_join = content_join
        self.row_padding = row_padding

    @classmethod
    def prep_input_arrays(cls, headers, data, header_spans):
        if headers is not None:
            if isinstance(headers, np.ndarray):
                if headers.ndim == 1: headers = headers[np.newaxis]
                header_cols = headers.shape[1]
            else:
                if isinstance(headers[0], (str, float, int, np.integer, np.floating)):
                    headers = [headers]
                header_cols = max(len(h) for h in headers)
            if header_spans is None:
                header_spans = []
            elif isinstance(header_spans[0], (int, np.integer)):
                header_spans = [header_spans]
            header_spans = header_spans + [
                [1] * header_cols
            ] * (len(headers) - len(header_spans))
        else:
            header_cols = 0
            header_spans = None

        if isinstance(data, np.ndarray):
            data_cols = data.shape[1]
        else:
            data_cols = max(len(a) for a in data)

        max_cols = max([header_cols, data_cols])

        data = [
            list(d) + [""]*(max_cols - len(d))
            for d in data
        ]

        if headers is not None:
            headers = [
                list(h) + [""]*(max_cols - len(h))
                for h in headers
            ]
            # header_spans = [
            #     list(h) + [1]*(max_cols - len(h))
            #     for h in header_spans
            # ]

        return headers, data, header_spans
    @classmethod
    def custom_formatter(cls, f):
        if isinstance(f, str):
            return "{:" + f + "}" if not f.startswith("{") else f
        elif hasattr(f, 'format'):
            return f
        elif isinstance(f, (list, tuple)):
            fmt_func = cls.custom_formatter(f[0])
            pad = f[1] if len(f) > 1 else ""
            def format_iterable(obj, fmt_func=fmt_func, pad=pad):
                return pad.join(fmt_func.format(o) for o in obj)
            format_iterable.format = format_iterable
            return format_iterable
        else:
            def format_func(obj, f=f):
                return f(obj)
            format_func.format = format_func
            return format_func
    @classmethod
    def resolve_formatters(cls, ncols, col_formats):
        if isinstance(col_formats, str):
            col_formats = [col_formats]
        col_formats = list(col_formats)
        pad_formats = col_formats * (ncols//len(col_formats)+1)

        return pad_formats[:ncols]
    @classmethod
    def prep_formatters(cls, formats):
        if isinstance(formats, str):
            formats = [formats]
        return [
            cls.custom_formatter(f)
            for f in formats
        ]

    @classmethod
    def format_tablular_data_columns(cls, data, formats, row_padding=None):
        ncols = len(data[0])
        return [
            [
                (
                    row_padding + formats[i].format(d[i])
                        if row_padding is not None and i == 0 else
                    formats[i].format(d[i])
                ) for d in data
            ]
            for i in range(ncols)
        ]

    @classmethod
    def align_left(cls, col, width):
        f_spec = "{:<"+str(width)+"}"
        return [f_spec.format(c) for c in col]
    @classmethod
    def align_right(cls, col, width):
        f_spec = "{:>" + str(width) + "}"
        return [f_spec.format(c) for c in col]

    @classmethod
    def align_center(cls, col, width):
        f_spec = "{:^" + str(width) + "}"
        return [f_spec.format(c) for c in col]

    @classmethod
    def align_dot(cls, col, width, dot='.'):
        dot_pos = [c.rfind(dot) for c in col]
        dot_pos = [
            # find position from right to pad everything to the same number of spaces after the decimal
            len(c) - dp if dp > 0 else -1
            for c,dp in zip(col, dot_pos)
        ]
        max_dot_pos = max(dot_pos)
        col = [
            c + " "*(max_dot_pos - dp)
            for c,dp in zip(col, dot_pos)
        ]
        new_width = max(len(c) for c in col)
        return cls.align_right(col, max(width, new_width))
    supported_alignments = {
            '<':'align_left',
            '>':'align_right',
            '^':'align_center',
            '.':'align_dot'
        }
    @classmethod
    def resolve_aligner(cls, alignment): return getattr(cls, cls.supported_alignments[alignment])
    @classmethod
    def align_column(cls,
                     header_data, cols_data, header_alignment, column_alignment, join_width,
                     header_widths
                     ):
        if header_data is not None:
            header_width = max(len(c) for c in header_data)
        else:
            header_width = 0

        col_widths = [
            max(len(c) for c in col)
            for col in cols_data
        ]
        cols_data = [
            cls.resolve_aligner(al)(c, w)
            for c, w, al in zip(cols_data, col_widths, column_alignment)
        ]
        col_widths = [
            max(len(c) for c in col)
            for col in cols_data
        ]


        max_width = sum(col_widths)
        width = max([header_width, max_width]) + join_width*max(len(col_widths)-1, 0)
        if header_data is not None:
            header_data = cls.resolve_aligner(header_alignment)(header_data, width)

            if len(col_widths) == 1:
                header_width = max(len(c) for c in header_data)
                width = max([header_width, max_width]) + join_width*max(len(col_widths)-1, 0)
                cols_data = [
                    cls.resolve_aligner(al)(c, width)
                    for c, al in zip(cols_data, column_alignment)
                ]

        # if format_data:
        # cols_data = [
        #     cls.resolve_aligner(al)(c, w)
        #     for c,w,al in zip(cols_data, col_widths, column_alignment)
        # ]
        return header_data, cols_data

    def format(self,
               headers_or_table, *table_data,
               header_format=None,
               header_spans=None,
               column_formats=None,
               column_alignments=None,
               header_alignments=None,
               column_join=None,
               row_join=None,
               header_column_join=None,
               header_row_join=None,
               separator=None,
               separator_lines=None,
               content_join=None,
               row_padding=None
               ):
        if len(table_data) == 0:
            headers = None
            table_data = headers_or_table
        elif len(table_data) > 1:
            raise ValueError("expected only one tabular data argument")
        else:
            headers = headers_or_table
            table_data = table_data[0]

        if headers is None: headers = self.headers
        if header_spans is None: header_spans = self.header_spans

        headers, table_data, header_spans = self.prep_input_arrays(headers, table_data, header_spans)

        if column_formats is None:
            if column_formats is None: column_formats = self.column_formats
        column_formats = self.resolve_formatters(len(table_data[0]), self.prep_formatters(column_formats))

        if column_alignments is None:
            if column_alignments is None: column_alignments = self.column_alignments
            if column_alignments is None: column_alignments = self.default_column_alignment
        if isinstance(column_alignments, str): column_alignments = [column_alignments]
        column_alignments = self.resolve_formatters(len(table_data[0]), column_alignments)

        for alignment in column_alignments:
            if alignment not in self.supported_alignments:
                raise ValueError("unknown alignment spec {}".format(alignment))

        if headers is not None:
            if header_alignments is None:
                if header_alignments is None: header_alignments = self.header_alignments
                if header_alignments is None: header_alignments = self.default_header_alignment
            if isinstance(header_alignments, str): header_alignments = [header_alignments]
            header_alignments = self.resolve_formatters(len(table_data[0]), header_alignments)

            for alignment in header_alignments:
                if alignment not in self.supported_alignments:
                    raise ValueError("unknown alignment spec {}".format(alignment))

        if headers is not None:
            if header_format is None: header_format = self.header_format
            if header_format is None: header_format = self.default_header_format
            header_format = self.resolve_formatters(len(headers[0]), self.prep_formatters(header_format))

        if headers is not None:
            header_columns = self.format_tablular_data_columns(headers, header_format)
        else:
            header_columns = None

        if row_padding is None: row_padding = self.row_padding
        if row_padding is None: row_padding = self.default_row_padding
        data_columns = self.format_tablular_data_columns(table_data, column_formats, row_padding)

        if column_join is None: column_join = self.column_join
        if column_join is None: column_join = self.default_column_join

        if headers is None:
            col_ = []
            sp = [1] * len(data_columns)
            split_cols = []
            split_alignments = []
            p = 0
            for s in sp:
                split_alignments.append(column_alignments[p:p + s])
                split_cols.append(data_columns[p:p + s])
                p += s

            for d,al in zip(split_cols, column_alignments):
                _, dc = self.align_column(
                    None, d,
                    None, al,
                    0, True
                )
                col_.extend(dc)

            data_columns = col_
        else:
            header_rows = []
            col_ = []

            join_width = len(column_join) if isinstance(column_join, str) else 0

            for i,sp in enumerate(header_spans):
                subcol = []

                header_subcol = [
                    [hc[i]]
                    for hc in header_columns
                ]

                split_cols = []
                split_alignments = []
                p = 0
                for s in sp:
                    split_alignments.append(column_alignments[p:p + s])
                    split_cols.append(data_columns[p:p + s])
                    p += s

                subrow = []
                # format_cols = i == 0
                for h,c,hal,cal in zip(
                        header_subcol, split_cols,
                        header_alignments, split_alignments
                ):
                    hc, dc = self.align_column(
                        h, c,
                        hal, cal,
                        join_width, True
                    )
                    subrow.extend(hc)
                    subcol.extend(dc)
                if len(col_) == 0:
                    col_ = subcol
                else:
                    col_ = [
                        c1 if len(c1[0]) >= len(c2[0]) else c2
                        for c1, c2 in zip(col_, subcol)
                    ]
            data_columns = col_

            for i, sp in enumerate(header_spans):
                header_subcol = [
                    [hc[i]]
                    for hc in header_columns
                ]

                split_cols = []
                split_alignments = []
                p = 0
                for s in sp:
                    split_alignments.append(column_alignments[p:p + s])
                    split_cols.append(data_columns[p:p + s])
                    p += s

                subrow = []
                # format_cols = i == 0
                for h, c, hal, cal in zip(
                        header_subcol, split_cols,
                        header_alignments, split_alignments
                ):
                    hc, dc = self.align_column(
                        h, c,
                        hal, cal,
                        join_width, True
                    )
                    subrow.extend(hc)
                header_rows.append(subrow)

        data_rows = [
            [dc[n] for dc in data_columns]
            for n in range(len(data_columns[0]))
        ]

        if row_join is None: row_join = self.row_join
        if row_join is None: row_join = self.default_row_join

        body_rows = [
            column_join.join(r) for r in data_rows
        ]
        body = row_join.join(body_rows)

        if headers is not None:
            if header_column_join is None: header_column_join = self.header_column_join
            if header_column_join is None: header_column_join = column_join

            header_rows = [header_column_join.join(r) for r in header_rows]

            if header_row_join is None: header_row_join = self.header_row_join
            if header_row_join is None: header_row_join = row_join

            if separator is None: separator = self.separator
            if separator is None: separator = self.default_separator
            if separator_lines is None: separator_lines = self.separator_lines

            sep_line = (separator * len(header_rows[0]))[:len(header_rows[0])]
            header_rows = header_rows + [sep_line for i in range(separator_lines)]
            header = header_row_join.join(header_rows)

            if content_join is None: content_join = self.content_join
            if content_join is None: content_join = row_join
            body = content_join.join([header, body])


        return body








"""
Module that provides a FiniteDifferenceDerivative class that does finite-difference derivatives
"""

from .FiniteDifferenceFunction import FiniteDifferenceFunction
from ...Parallelizers import Parallelizer, SerialNonParallelizer
from ...Scaffolding import Logger, NullLogger
import numpy as np, itertools as it, scipy.sparse as sparse

__all__ = [
    'FiniteDifferenceDerivative'
]

class FiniteDifferenceDerivative:
    """
    Provides derivatives for a function (scalar or vector valued).
    Can be indexed into or the entire tensor of derivatives may be requested.
    The potential for optimization undoubtedly exists, but the idea is to provide as _simple_ an interface as possible.
    Robustification needs to be done, but is currently used in `CoordinateSystem.jacobian` to good effect.
    """
    __props__ = (
        'function_shape',
        'displacement_function',
        'prep',
        'lazy',
        'mesh_spacing',
        'stencil',
        'accuracy',
        'cache_evaluations',
        'parallelizer',
        'logger'
    )
    def __init__(self,
                 f,
                 function_shape=(0, 0),
                 parallelizer=None,
                 logger=None,
                 **fd_opts
                 ):
        """
        :param f: the function we would like to take derivatives of
        :type f: FunctionSpec | callable
        :param function_shape: the shape of the function we'd like to take the derivatives of
        :type function_shape: Iterable[Iterable[int] | int] | None
        :param fd_opts: the options to pass to the finite difference function
        :type fd_opts:
        """
        if not isinstance(f, FunctionSpec):
            f = FunctionSpec(f, *function_shape)
        self.f = f
        self.parallelizer=parallelizer
        self.logger=logger
        # if parallelizer is not None:
        #     raise NotImplementedError("need to figure out how I want to work parallelism in")
        self._fd_opts = fd_opts

    def __call__(self, *args, **opts):
        return self.derivatives(*args, **opts)

    def derivatives(self,
                    center,
                    displacement_function=None,
                    prep=None,
                    lazy=None,
                    mesh_spacing=None,
                    **fd_opts
                    ):
        """
        Generates a differencer object that can be used to get derivs however your little heart desires

        :param center: the center point around which to generate differences
        :type center: np.ndarray
        :param displacement_function:
        :type displacement_function:
        :param mesh_spacing:
        :type mesh_spacing:
        :param prep:
        :type prep:
        :param fd_opts:
        :type fd_opts:
        :return:
        :rtype:
        """

        f = self.f
        fd_opts = dict(self._fd_opts, **fd_opts)
        for k, v, d in zip(
                ("displacement_function", "prep", "lazy", "mesh_spacing"),
                (displacement_function, prep, lazy, mesh_spacing),
                (None, None, False, .001)
        ):
            if v is not None:
                fd_opts[k] = v
            elif k not in fd_opts:
                fd_opts[k] = d

        return DerivativeGenerator(
            f,
            center,
            parallelizer=self.parallelizer,
            logger=self.logger,
            **fd_opts
        )

class FunctionSpec:
    """
    Defines a general spec that specifies a function, what it takes as coordinate inputs, and what the dimensions of what it outputs
    """

    def __init__(self, f, input_shape, output_shape):
        """

        :param f:
        :type f:
        :param input_shape: the shape of the array that should be passed in
        :type input_shape:
        :param output_shape:
        :type output_shape:
        """
        self.f = f
        self.in_shape = input_shape
        self.output_shape = output_shape
    def __call__(self, *args, **kwargs):
        return self.f(*args, **kwargs)

class DerivativeGenerator:
    """A that generates specified derivatives, currently by FD but can be generalized out to do it other ways

    The lower-level class that takes the entire high-level spec and converts it into derivatives.
    Allows indexing for memory efficiency
    """
    __props__ = (
        'displacement_function',
        'prep',
        'lazy',
        'mesh_spacing',
        'stencil',
        'accuracy',
        'cache_evaluations',
        'parallelizer',
        'logger'
    )
    def __init__(self,
                 f_spec,
                 center,
                 displacement_function=None,
                 prep=None,
                 lazy=False,
                 mesh_spacing=.001,
                 cache_evaluations=True,
                 parallelizer=None,
                 logger=None,
                 **fd_opts
                 ):
        """

        :param f_spec: The function to use when evaluating at the various displacements
        :type f_spec: FunctionSpec | function
        :param center: The position to evaluate the derivatives at
        :type center: np.ndarray
        :param displacement_function: The function to generate displacements
        :type displacement_function: FunctionSpec | function | None
        :param prep:
        :type prep: function | None
        :param mesh_spacing:
        :type mesh_spacing: float
        :param symmetric:
        :type symmetric:
        :param cache_evaluations: whether or not to cache function evaluations for reuse in other derivatives
        :type cache_evaluations: bool
        :param fd_opts:
        :type fd_opts:
        """

        # set up the coordinate parameters
        if isinstance(center, (int, float, np.integer, np.floating)):
            center = [center]
        center = np.asarray(center)

        input_shape = f_spec.in_shape if isinstance(f_spec, FunctionSpec) else 0
        # determine if center is multiple configurations or just one
        input_dims = (0 if isinstance(input_shape, (int, np.integer)) else len(input_shape))
        multiconfig = len(center.shape) > input_dims
        if not multiconfig:
            coords = np.array([center])
        else:
            coords = np.asarray(center)
            if len(coords.shape) == 1:
                coords = coords.reshape(coords.shape + (1,))
        if input_dims == 0:
            input_dims = 1

        # determine up the data shapes
        configs_dims = coords.shape[:-input_dims]
        coord_shape = tuple(coords.shape[-input_dims:])
        # coord_shape = tuple(shape if isinstance(input_shape, (int, np.integer)) else input_shape)

        # figure out how many dimensions we will have flattened for a multiconfiguration state
        flattened_dims = None
        if multiconfig:
            flattened_dims = configs_dims
            num_flattened = np.prod(flattened_dims)
            # I think by being smarter I could have avoided this reshape and avoided a bunch of hassle
            coords = coords.reshape((num_flattened, ) + coord_shape)

        # configure the function for handling displacements
        if displacement_function is None:
            displacement_function = self._default_displace

        # define a function to prep the system
        if prep is None:
            prep = self._default_prep

        self.f = f_spec
        self.config_shape = configs_dims
        self.coord_shape = coord_shape

        self.center = coords
        self.multiconfig = multiconfig
        self.prep = prep

        self.lazy = lazy

        self.fd_opts = fd_opts
        self._flattend_dims = flattened_dims
        self.displacement_function = displacement_function
        self.mesh_spacing = mesh_spacing
        self._fdfs = {}

        self.cache_evaluations = cache_evaluations
        self._cache = {}

        self.parallelizer = parallelizer
        self.logger = NullLogger() if logger is None else logger

    def __getstate__(self):
        state = {}
        for k in self.__dict__:
            if k != '_cache':
                state[k] = self.__dict__[k]
        state['_cache'] = None
        return state

    def __setstate__(self, state):
        for k in state:
            self.__dict__[k] = state[k]
        self._cache = {}

    @staticmethod
    def _default_displace(c, a):
        if not isinstance(a, (int, float, np.integer, np.floating)):
            a = a[c % len(a)]
        return a

    @staticmethod
    def _default_prep(c, a, b):
        return (a, b)

    def _get_fdf(self, ci, mesh_spacing):
        # create the different types of finite differences we'll compute for the different coordinates of interest
        dorder = self._dorder(ci)

        try:
            fdf = self._fdfs[(dorder, mesh_spacing)]
        except KeyError:
            fdf = FiniteDifferenceFunction.regular_difference(
                dorder,
                end_point_accuracy=0,
                only_core=False,
                only_center=True,
                contract=True,
                mesh_spacing=mesh_spacing,
                **self.fd_opts
            )
            self._fdfs[(dorder, mesh_spacing)] = fdf

        stencil_widths = tuple(len(cf[1]) if cf is not None else cf for cf in fdf.weights)
        stencil_shapes = tuple(w[1] if w is not None else w for w in fdf.widths)
        if any(x != sum(y) for x,y in zip(stencil_widths, stencil_shapes)):
            raise ValueError("weights {} don't correspond to shapes {}...".format(
                fdf.weights,
                fdf.widths
            ))
        finite_difference = fdf#.get_FDF(shape = stencil_widths)

        return stencil_widths, stencil_shapes, finite_difference, dorder

    def _coord_index(self, coord):
        """
        Converts the coordinate spec into a linear index

        :param coord:
        :type coord:
        :return:
        :rtype:
        """
        # gotta re-linearize coord so we can use it to index into like fdfs and stuff...
        if coord.ndim == 1:
            raveled = np.ravel_multi_index(coord, self.coord_shape)
        else:
            raveled = [ np.ravel_multi_index(c, self.coord_shape) for c in coord ]
        return raveled

    def _dorder(self, raveled):
        """Computes the derivative order requested from the coordinate indices requested

        :param raveled:
        :type raveled:
        :return:
        :rtype:
        """
        unique, counts = np.unique(raveled, return_counts=True)
        pairs = np.array([unique, counts]).T
        sorting = np.argsort(pairs[:, 0])
        orders = pairs[sorting, 1]
        return tuple(orders)

    def get_displacement(self, coord, mesh_spacing=None):
        """
        Computes the displacement for the passed mesh spacing

        :param coord:
        :type coord:
        :param mesh_spacing:
        :type mesh_spacing:
        :return:
        :rtype:
        """
        if mesh_spacing is None:
            mesh_spacing = self.mesh_spacing
        displacement = self.displacement_function(self._coord_index(coord)[0], mesh_spacing)
        if isinstance(displacement, (float, np.floating)):
            displacement = np.full(self.coord_shape, displacement)
        elif displacement.shape == self.coord_shape[-1:]:
            displacement = np.broadcast_to(displacement, self.coord_shape)

        return displacement

    def _build_displacement_array(self, coord, stencil_shapes, stencil_widths, displacement, use_sparse=False):

        # not too worried about looping over coordinates since the number of loops will be like max in the small hundreds
        num_displacements = np.prod(stencil_widths)
        displacement_shape = (num_displacements,) + displacement.shape

        if use_sparse:
            displacements = sparse.lil_matrix(displacement_shape)
        else:
            displacements = np.zeros(displacement_shape)  # fuck ton of zeros
        base_roll = tuple(np.arange(len(stencil_widths)))

        coord = np.unique(coord, axis=0)

        mesh_spacings = []
        for i, sc in enumerate(zip(stencil_shapes, coord)):
            stencil_shape, c = sc
            # coord can be like ((1, 1), (1, 2)) or ((0,), (1,)) in which case we have a 2D derivative
            # creates single displacement matrix
            # it can also just be like ((0,)) in which case we're working in 1D
            ctup = tuple(c)

            # pulls the appropriate displacement for this coordinate specifically
            disp = displacement[ctup]
            mesh_spacings.append(disp)

            # generate a full tensor of displacements by scaling the
            # initial displacement by each of the number of steps to be taken
            steps = np.arange(-(stencil_shape[0] - 1), stencil_shape[1] + 1)
            disp = disp * steps

            # pad this full displacement so it has enough 1s in it to be broadcasted to the stencil
            # shape
            full_disp = np.reshape(disp, disp.shape + (1,) * (len(stencil_widths) - disp.ndim))
            # keep moving the axis for the displacement
            roll = np.roll(base_roll, i)
            full_disp = full_disp.transpose(roll)
            # print(i, base_roll, full_disp.shape, stencil_widths)
            to_set = np.broadcast_to(full_disp, stencil_widths).flatten()

            # in dimension 1 we want to have this repeat over the slowest moving indices, then the ones after those
            # then after those, etc.
            idx = (...,) + ctup
            displacements[idx] = to_set

        return displacements, displacement_shape, mesh_spacings

    def _get_displaced_coords(self, coord, stencil_widths, stencil_shapes, use_sparse = False):
        """
        Provides enough displacements of along `coord` to satisfy `stencil_widths` and `stencil_shapes`
        Does this by generating a big tensor of zeros, assigning parts of this for each displacement, then adding that

        :param coord:
        :type coord:
        :param stencil_widths:
        :type stencil_widths:
        :param stencil_shapes:
        :type stencil_shapes:
        :return:
        :rtype:
        """

        coords = self.center
        displacement = self.get_displacement(coord)

        displacements, displacement_shape, mesh_spacings = self._build_displacement_array(
            coord, stencil_shapes, stencil_widths, displacement, use_sparse=use_sparse
        )

        # then we broadcast this up to the total number of structures we have
        full_target_shape = self.config_shape + displacement_shape
        coords_expanded = coords.reshape(
            self.config_shape
            + (1,) * (len(displacement_shape) - len(self.coord_shape))
            + self.coord_shape
        )
        displacements = displacements.reshape(
            (1,) * len(self.config_shape)
            + displacement_shape
        )
        displaced_coords = (
            np.broadcast_to(coords_expanded, full_target_shape) +
            np.broadcast_to(displacements, full_target_shape)
        )

        disp_spec = (
            tuple(tuple(x for x in c) for c in coord),
            stencil_widths,
            displacement_shape
        )

        return disp_spec, displacements, displaced_coords, mesh_spacings

    def _get_fd_data(self, specs):
        """
        Takes the specs and returns a generator that will create the appropriate derivatives along each coordinate
        I should add an optimization that allows displacements along single coordinates to happen fast...

        :param specs:
        :type specs:
        :param stencil_widths:
        :type stencil_widths:
        :param stencil_shapes:
        :type stencil_shapes:
        :return:
        :rtype:
        """

        for spec in specs: # dumb for now but allows me to pick some optimal ordering in the future
            # print(" >>>>>", spec)
            ci = self._coord_index(spec)
            h = self.displacement_function(ci[0], self.mesh_spacing)
            fd_data = self._get_fdf(ci, h)
            # print("   ::", ci, fd_data[0], fd_data[1])
            disp_data = self._get_displaced_coords(spec, fd_data[0], fd_data[1])
            # print("   ::", disp_data[0], disp_data[1].shape)

            yield spec, disp_data, fd_data

    @staticmethod
    def _get_diff(c, disp):
        """
        Gets the mesh spacing along coordinate c from the displaced coordinates...except there's basically no
        reason for this when I could instead _just save the number_

        :param c:
        :type c:
        :param disp:
        :type disp:
        :return:
        :rtype:
        """

        # diffs = np.abs(np.diff()) # why do it like this...?????
        # return np.sort(diffs)[-1]  # get the only diff > 0 (assumes we have a regular grid)
        subgrid = np.unique(disp[(...,) + tuple(c)])
        return abs(subgrid[1] - subgrid[0])

    def _get_single_deriv(self, spec, disp_data, fd_data, return_coords):

        if self.cache_evaluations:
            # we assume that evaluating the hash of the array will be faster
            # than evaluating f itself
            # this might not aways be the case,
            # but if it fails to be so that's likely not a cause for concern because it
            # means the overall process is already fast
            cache_key = disp_data[0]
            cached = cache_key in self._cache
            # if uncached:
            #     if not hasattr(self, "_cache_fails"):
            #         self._cache_fails = 0
            #     else:
            #         self._cache_fails += 1
            #     print(spec, disp_data[0], disp_data[1].shape)
            #     if self._cache_fails > 20:
            #         raise Exception("wat")
        else:
            cache_key = None
            cached = False


        f = self.f
        # Disp spec is a key
        # displacements tell us the flattend shape of the displacements
        # displaced coords are the proper coords to eval
        disp_spec, displacements, displaced_coords, mesh_spacings = disp_data
        stencil_widths, stencil_shapes, finite_difference, dorder = fd_data

        # TODO: sit down and add comments to this compact implementation...

        if cached:
            (disp, fvals) = self._cache[cache_key]
            # print("oooh?", cache_key)
        else:
            config_shape = self.config_shape
            # tries to unravel the displacements into a form that
            # mimics the shape of the FD product grid
            cdim = len(config_shape)
            dcoords = np.moveaxis(displaced_coords, cdim, 0)
            function_values = f(dcoords)
            function_values = np.moveaxis(function_values, 0, cdim)
            disp, fvals = self.prep(spec, displacements, function_values)

            # TODO: handle stuff like dipoles where we have an x, y, z component each of which should be handled separately...
            #       this might actually be handled naturally, though? I'm actually pretty hopeful it will be...
            #       UPDATE: it's not, I should fix this

            # we now need to reformat the fvals so that they respect the stencil_shapes
            # not sure where the extra 1 is coming from here...?
            # I think maybe from the transpose call...?
            # Actually no that just puts the displacement back in the right spot...
            #   seems that it comes out of the way the coordinates get fed in/come out...
            #   the tested stuff all has an extra '1' because of the way the displacement
            #   gets generated
            out_shape = fvals.shape[1+len(self.config_shape):]
            out_dim = 1 if isinstance(out_shape, (int, np.integer)) else len(out_shape)
            fvals_shape = self.config_shape + stencil_widths
            if out_dim > 0:
                fvals_shape = fvals_shape + fvals.shape[-out_dim:]
            fvals = fvals.reshape(fvals_shape)

        h = [self._get_diff(c, disp) for c in spec] if mesh_spacings is None else mesh_spacings

        derivs = finite_difference(fvals, axes=len(self.config_shape), mesh_spacing=h)

        if self.cache_evaluations and not cached:
            self._cache[cache_key] = (disp, fvals)

        if return_coords:
            return displaced_coords, derivs
        else:
            return derivs

    def _spec_derivs(self, specs, return_coords=False):
        """
        Computes a specific derivative
        The derivative is specified by repeating a coordinate the number of times we'd like its derivative:
            e.g. `[[0], [0]]` for $dx_1^2$ or `[[0], [N]]` for $dx_1dx_N$ if we have a function $f:R_N \rightarrow R$
                or `[[0, 1], [1, 0]]` for $dA_1_2dA_2_1$ if we have a function $f:(R_M, R_N) \rightarrow R$

        :param spec: the derivative to take as specified by the given indices
        :type spec: Iterable[Iterable[int]]
        :return:
        :rtype:
        """

        for spec, disp_data, fd_data in self._get_fd_data(specs):
            yield self._get_single_deriv(spec, disp_data, fd_data, return_coords)

    def _get_specs(self, order, pos=(), coordinates=None):
        """
        We compute the positions defined by the total order of the derivative as they would show up in the total tensor
        If a given block of derivatives is specified **[NOTE: I didn't finish this docstring and have no idea what it was supposed to say...]**

        :param order:
        :type order: int
        :param pos:
        :type pos: Iterable[int]
        :param coordinates:
        :type coordinates:
        :return:
        :rtype:
        """
        if coordinates is None:
            coordinates = np.arange(np.prod(self.coord_shape))
        elif not isinstance(coordinates[0], (int, np.integer)):
            coordinates = self._fidx(coordinates)

        # print(coordinates, self._fidx, self.coord_shape)

        pos = [slice(None, None, None) if a is None else a for a in pos]
        if len(pos) == order and all(isinstance(i, (int, np.integer)) for i in pos):
            # we're just asking for a single derivative (I think?)
            # basically when we have a derivative of order `n` and have something like `[i_1, i_2, ..., i_n]`
            # as the `pos`
            specs = [tuple(coordinates[c] for c in pos)]
        else:
            # we need to broadcast our specs
            sub_specs = [ tuple(coordinates[a]) if not isinstance(a, (int, np.integer)) else (coordinates[a],) for a in pos]
            sub_specs = sub_specs + [coordinates for i in range(order - len(pos))]
            unique = set()
            def test(p):
                # assumes derivative tensors are totally symmetric (good assumption in science)
                # filters out permutations by whether or not they are equivalent under sorting
                s = tuple(sorted(p))
                if s in unique:
                    return False
                else:
                    unique.add(s)
                    return True
            # we take the product space of the provided specs, since each spec targets a specific dimension
            # this is basically numpy broadcasting
            # then we filter by symmetry
            specs = [p for p in it.product(*sub_specs) if test(p)]

        return [self._idx(s) for s in specs], specs

    def _parallel_derivs(self, specs=None, parallelizer=None):
        # if main_kwargs is not None:
        #     specs = main_kwargs['specs']
        # else:
        #     specs = None
        specs = parallelizer.scatter(specs)
        derivs = np.array([x for x in self._spec_derivs(specs)])
        # parallelizer.print(derivs)
        derivs = parallelizer.gather(derivs)
        if parallelizer.on_main:
            if not isinstance(derivs, np.ndarray):
                derivs = np.concatenate(derivs, axis=0)
        return derivs
    def compute_derivatives(self, order, pos=(), coordinates=None, lazy=None, parallelizer=None):
        """
        Computes the derivatives up to `order` filtered by `pos` over the `coordinates`

        :param order: the maximum total order of the derivatives to calculate
        :type order: int | Iterable[int]
        :param pos: the positions to filter on
        :type pos:
        :param coordinates: the coordinates to compute the tensor of derivatives over
        :type coordinates: Iterable[int] | None
        :param lazy: whether or not to return a generator
        :type lazy: bool | None
        :return:
        :rtype:
        """

        if isinstance(order, (int, np.integer)):
            single = True
            order = [order]
        else:
            single = False

        res = [None]*len(order)
        def lazy_derivs(ders):
            for d in ders:
                yield d

        lazy = self.lazy if lazy is None else lazy
        if lazy:
            for which, o in enumerate(order):
                specs, raw = self._get_specs(o, pos, coordinates)
                derivs = self._spec_derivs(specs)
                res[which] = lazy_derivs(derivs)
        else:
            if parallelizer is None:
                parallelizer = self.parallelizer
            par = Parallelizer.lookup(parallelizer)
            if isinstance(par, SerialNonParallelizer):
                for which, o in enumerate(order):
                    specs, raw = self._get_specs(o, pos, coordinates)
                    derivs = np.array([s for s in self._spec_derivs(specs)])
                    res[which] = derivs
            else:
                with par:
                    for which, o in enumerate(order):
                        specs, raw = self._get_specs(o, pos, coordinates)
                        comm = None if len(specs)>=par.nprocs else list(range(len(specs)))
                        derivs = par.run(self._parallel_derivs, main_kwargs={'specs':specs},
                                         comm=comm
                                         )
                        res[which] = derivs
        if single:
            res = res[0]
        return res

    def derivative_tensor(self, order, pos=(), coordinates=None, parallelizer=None, logger=None):
        """
        Computes a given derivative tensor

        :param order:
        :type order: int | Iterable[int]
        :param pos:
        :type pos:
        :param coordinates:
        :type coordinates:
        :return:
        :rtype:
        """

        if parallelizer is None:
            parallelizer = self.parallelizer

        if logger is None:
            logger = self.logger

        if isinstance(order, (int, np.integer)):
            single = True
            order = [order]
        else:
            single = False

        res = [None]*len(order)

        with logger.block(tag="getting derivative tensors at order {}".format(order)):

            if parallelizer is None:
                parallelizer = self.parallelizer
            par = Parallelizer.lookup(parallelizer)
            with par:
                for which, o in enumerate(order):
                    pos = [slice(None, None, None) if a is None else a for a in pos]
                    if coordinates is None:
                        coordinates = np.arange(np.prod(self.coord_shape))
                    elif not isinstance(coordinates[0], (int, np.integer)):
                        coordinates = self._fidx(coordinates)

                    sub_specs = [coordinates[a] if not isinstance(a, (int, np.integer)) else (coordinates[a],) for a in pos]
                    sub_specs = sub_specs + [coordinates for i in range(o - len(pos))]
                    specs, raw = self._get_specs(o, pos, coordinates)

                    if isinstance(par, SerialNonParallelizer):
                        logger.log_print("evaluating {nspec} derivatives", nspec=len(specs))
                        derivs = np.array([s for s in self._spec_derivs(specs)])
                    else:
                        comm = None if len(specs) >= par.nprocs else list(range(len(specs)))
                        logger.log_print("evaluating {nspec} derivatives over {nproc} processors", nspec=len(specs), nproc=par.nprocs)
                        derivs = par.run(self._parallel_derivs, main_kwargs={'specs':specs},
                                         comm=comm
                                         )
                    # derivs = self._spec_derivs(specs)

                    d_tensor = None
                    # apply symmetry
                    for s, d in zip(raw, derivs):
                        if d_tensor is None:
                            d_tensor = np.ones(tuple(len(s) for s in sub_specs) + d.shape)

                        for p in it.permutations(s, len(s)):
                            try:
                                d_tensor[p] = d
                            except IndexError:
                                pass

                    res[which] = d_tensor

        if single:
            res = res[0]
        return res

    def _idx(self, c, coord_shape = None):
        if coord_shape is None:
            coord_shape = self.coord_shape
        return np.array(np.unravel_index(np.array(c), coord_shape)).T
    def _fidx(self, c, coord_shape = None):
        if coord_shape is None:
            coord_shape = self.coord_shape
        return np.ravel_multi_index(c, coord_shape)

    def __getitem__(self, item):
        """
        Returns a single derivative or block of derivatives, depending on how `item` is defined

        :param item: position spec
        :type item: int | tuple[int] | slice | tuple[int | slice]
        :return: derivs
        :rtype: float | np.ndarray
        """
        if isinstance(item, (int, np.integer)):
            item = (item,) # though the branches might differ...
        elif isinstance(item, slice):
            item = (item,)
        if any(not isinstance(i, (int, np.integer)) for i in item):
            return self.derivative_tensor(len(item), item)
        else:
            d = self.compute_derivatives(len(item), item, lazy=False)
            return d[0]

